/*
 * Worldvisions Weaver Software:
 *   Copyright (C) 1997-2002 Net Integration Technologies, Inc.
 *
 * Definition of the WvFastForward class, which listens for connections on a 
 * given set of IP addresses and ports, and redirects them to other locations,
 * based on data in the config file.  For example:
 * 
 * 	[Fast Forward]
 *	aminal:4250 = insight:25
 *
 * will (if running on aminal) accept connections to port 4250 of aminal, and
 * pipe them to insight's sendmail port.
 *
 * If 0.0.0.0 is specified as the "source" IP address, then WvFastForward will
 * listen for connections to port 4250 (in this case) of whatever IP addresses
 * it's able to.
 *
 * Fast Forward also forwards UDP packets.  Currently there is no way to 
 * specify whether you want TCP or UDP -- it always does both.  Fix that.
 *
 */

#include "wvfastfwd.h"
#include "wvver.h"

#include <stdlib.h>
#include <time.h>
#include <netdb.h>		/* getservbyname */

bool make_port_addr( WvIPPortAddr& addr, const char * str, char ** err )
/**********************************************************************/
// Takes a string that might look like "aminal:4250" and resolves the hostname,
// dumping the result into addr.
{
    char *	     p = strchr( str, ':' );
    WvResolver       dns;
    const WvIPAddr * tmp;
    int		     result;
    static char	     err_buf[ 1024 ];
    int              port;
    struct servent   *s;

    if( p == NULL ) {
    	sprintf( err_buf, "Port number not specified in %.128s\n", str );
    	*err = err_buf;
    	return( false );
    }

    *(p++) = '\0';	// Remember to put the colon back later!

    // Try to convert a service name to port number.
    // This would be a lot easier if we knew whether we were dealing
    // with tcp or udp. For now let's check tcp first then udp.
    // Are there any services which the tcp and udp port differ?
    if((s = getservbyname(p, "tcp")) != NULL) {
	port = ntohs( s->s_port );
    } else if((s = getservbyname(p, "udp")) !=NULL ) {
	port = ntohs( s->s_port );
    } else {
	port = atoi( p );
    }

    result = -1;
    while( result < 0 )
    	result = dns.findaddr( 100, str, &tmp );

    if( result == 0 ) {
    	sprintf( err_buf, "DNS: %.128s does not exist.  Ignoring.\n", str );
    	*err = err_buf;
    	*(p-1) = ':';	// So config data stays sane.
    	return( false );
    }

    addr = WvIPPortAddr( *tmp, port );

    *(p-1) = ':';	// So config data stays sane.
    *err = NULL;
    return( true );
}


///////////////////////////////////////////////////////////
// WvFastForward public functions
///////////////////////////////////////////////////////////

WvFastForward::WvFastForward( WvConf& conf )
/******************************************/
: log( "Fast Forward", WvLog::Debug ), err( log.split( WvLog::Error ) )
{
    cfg = &conf;
    broken_bind = false;
    shutting_down = false;

    log( WvLog::Info, "Fast Forward: proxy server version "
    	 FASTFORWARD_VER_STRING "\n" );
    
    append( &listeners, false, "listeners" );
    append( &tcp_streams, false, "tcp_streams" );
    append( &udp_streams, false, "udp_streams" );
}

void WvFastForward::clean_up()
/****************************/
{
    // If a stream has been inactive for more than TCP_XMIT_TIMEOUT seconds,
    // kill it.
    WvStreamList::Iter	iter( tcp_streams );

    for( iter.rewind(); iter.next(); ) {
    	WvFF_TCPConn& conn( (WvFF_TCPConn &) iter() );
	
    	if( time( NULL ) - conn.last_xmit 
	   > cfg->getint("Fast Forward", "Idle Timeout", TCP_XMIT_TIMEOUT )) {
    	    log( "Connection to %s idle too long.  Disconnecting.\n",
    	    	 *conn.src() );

	    if( conn.remote_conn )
		tcp_streams.unlink( conn.remote_conn );
    	    tcp_streams.unlink( &conn );

	    iter.rewind();
	}
    }
}

// Cancel all listener sockets, in anticipation of a reload_config() soon.
void WvFastForward::disable()
/***************************/
{
    // first, process all pending requests
    while (listeners.select(0, true, false))
	listeners.callback();
    
    // now clear the listeners list
    listeners.zap();
}

void WvFastForward::reload_config()
/*********************************/
// This function is used to tell WvFastForward that it should reload the
// pertinent info from the configuration file.  We need to shut down all the
// TCP listeners and then create new ones.
// NOTE: This should _NOT_ shut down any connections already in progress.
{
    // First we need to wipe out the listeners and the mapping list.
    disable();

    // Now we can just run start_listeners() again.
    start_listeners();
}

bool WvFastForward::too_many_conns() const
/****************************************/
{
    int in_use;
    int max;

    in_use = WvFF_TCPConn::num_conns();
    max    = 2 * cfg->getint( "Fast Forward", "Maximum Connections", 200 );
    // We have to multiply max by 2, since there are two WvFF_TCPConns for
    // every forwarding session...

    return( in_use >= max );
}


///////////////////////////////////////////////////////////
// WvFastForward private functions
///////////////////////////////////////////////////////////

bool WvFastForward::start_tcp_listener( const WvIPPortAddr& src,
					const char *	    dest,
					bool		    transproxy )
/**********************************************************************/
// For transproxy:
// If we're given something like "0.0.0.0:80 = transproxy" we must generate
// a random port number to bind to and pass to firewall.add_redir()...
{
    ipalias.add( src );

    WvIPPortAddr new_src = src;
    if( transproxy )
    	new_src.port = 0;	// WvTCPListener's constructor generates a 
    				// random number...

    WvFF_TCPListener * sock = new WvFF_TCPListener( new_src, this, dest, 
     						    transproxy );

    if( sock->isok() == false ) {
	log( WvLog::Warning, "Can't listen on TCP %s: %s\n",
	     new_src, sock->errstr() );
    	delete sock;
    	return( false );
    } else {
    	listeners.append( sock, true, "ff tcp listener" );

    	if( transproxy ) {
	    // Make sure the random port number gets displayed and send to
	    // add_redir()...
	    new_src.port = ( (WvIPPortAddr *)( sock->src() ) ) ->port;
	    firewall.add_redir( src, new_src.port );

    	    log( "Transproxy TCP from %s %s\n", new_src,
    	    	 ( dest && *dest ) ? WvString( "(to %s for loops).", dest )
    	    	                   : WvString( "(discarding loops)." ) );
	} else {
    	    log( "Forwarding TCP from %s to %s.\n", src, dest );
    	    firewall.add_port( src );
    	}

    	return( true );
    }
}

bool WvFastForward::start_udp_listener( const WvIPPortAddr& src,
					const char * 	    dest )
/****************************************************************/
{
    ipalias.add( src );

    WvFF_UDPListener * sock = new WvFF_UDPListener( src, this, dest );

    if( sock->isok() == false ) {
    	delete sock;
    	return( false );
    } else {
    	listeners.append( sock, true, "ff udp listener" );
    	log( "Forwarding UDP from %s to %s.\n", src, dest );

    	firewall.add_port( src );
    	return( true );
    }
}

bool WvFastForward::start_listeners()
/***********************************/
// We must read the config file, and create a TCP and UDP listener for each 
// port that we should be listening on.
{
    WvConfigSection *	sect = (*cfg)[ "Fast Forward" ];

    log( WvLog::Debug1, "Restarting listeners.\n" );

    if( sect == NULL ) {
    	err( "Config file needs a \"Fast Forward\" section.\n" );
    	return( false );
    }

    WvConfigEntryList::Iter	iter( *sect );

    firewall.zap();
    ipalias.start_edit();

    broken_bind = false;
    for( iter.rewind(); iter.next(); ) {
    	WvIPPortAddr	addr;
    	WvConfigEntry&  ent = *iter;
    	char *		err_buf;

    	if( strchr( ent.name, ' ' ) && !strstr( ent.name, " transproxy" ) )
    	    continue;	// fast forward option -- not a redirection!

	if( make_port_addr( addr, ent.name, &err_buf ) == false ) {
	    err( err_buf );
	    continue;
	}

	const char * tcp_dest   = ent.value;
	bool	     transproxy = false;
	bool	     do_udp     = true;
	if( !strncmp( ent.value, "transproxy", 10 ) ) {
	    transproxy = true;
	    tcp_dest = ent.value + 10;
	    do_udp = false;
	}

	bool ok1 =           start_tcp_listener( addr, tcp_dest, transproxy );
	bool ok2 = do_udp && start_udp_listener( addr, ent.value );
	bool ok3, ok4, bind_trouble;

	if( !ok1 || !ok2 ) {
	    bind_trouble = true;

	    if( WvIPAddr(addr) == WvIPAddr() ) {
	    	// local address 0.0.0.0
	    	WvInterfaceDict		interfaces;
	    	WvInterfaceDict::Iter	i( interfaces );

	    	bind_trouble = true;

	    	for( i.rewind(); i.next(); ) {
	    	    WvInterface& ifc = *i;

	    	    if( !ifc.valid && !ifc.isup() )
	    	    	continue;

	    	    WvIPPortAddr src( ifc.ipaddr(), addr.port );

	    	    ok3 = ok1 ||           start_tcp_listener( src, tcp_dest, 
							       transproxy );
	    	    ok4 = ok2 || do_udp && start_udp_listener( src, ent.value );

	    	    if( ok3 || ok4 )
	    	    	bind_trouble = false;
	    	}
	    }

	    if( !ok1 && !ok2 && bind_trouble ) {
	    	err( "Cannot bind to %s.  Ignored.\n", addr );
	    	broken_bind = true;
	    }
	}
    }

    ipalias.done_edit();

    if( !listeners.count() ) {
    	log( WvLog::Debug2, "Nothing to do.\n" );
    	return( false );
    }

    return( true );
}
