/*
 * Worldvisions Weaver Software:
 *   Copyright (C) 1997-2002 Net Integration Technologies, Inc.
 *
 * The TCP-related parts of Fast Forward.
 *
 */

#include "wvfftcp.h"

#include <time.h>
#include <sys/socket.h>


///////////////////////////////////////////////////////////
// WvFF_TCPListener stuff
///////////////////////////////////////////////////////////

WvFF_TCPListener::WvFF_TCPListener( const WvIPPortAddr& l_port,
				    WvFastForward *	forward,
				    const WvString& 	dest,
				    bool		transparent )
/*******************************************************************/
: WvTCPListener( l_port ), fwd( *forward ), dest_str( dest ),
    transproxy( transparent )
{
}


WvFF_TCPConn * WvFF_TCPListener::accept_ff()
/******************************************/
{
    struct sockaddr_in sin;
    size_t len = sizeof( sin );
    int    new_fd;
    WvFF_TCPConn * ret;

    // FIXME!  WvFF_TCPConn should be a WvStreamClone of a WvTCPConn.
    // This kind of direct access to ::accept() will bite us sooner or
    // later...
    new_fd = ::accept( getfd(), (struct sockaddr *) &sin, &len );
    ret = new WvFF_TCPConn( new_fd, WvIPPortAddr( &sin ) );
    return( ret );
}

void WvFF_TCPListener::execute()
/******************************/
{
    WvTCPListener::execute();
    
    WvFF_TCPConn * origin = accept_ff();

    if( transproxy )
    	// If this was a transparent proxy, we need to forward the connection
    	// to the LOCAL address of the origin... (since we're pretending to
    	// be the destination.)
    	fwd.log( "Transparent proxy from %s to %s...\n",
    			*origin->src(), origin->localaddr() );
    else
    	// If it WASN'T a transparent proxy, it's pretty easy, just forward
    	// it to the listener's stored destination.
	fwd.log( "Connection from %s to %s...\n",
    			*origin->src(), *src() );


    // If there are too many connections open, close this new one!
    if( fwd.too_many_conns() ) {
    	fwd.err( "Not forwarding -- too many connections open!\n" );
	delete origin;
    	return;
    }

    // Must now start the outgoing connection, after we find who it should
    // go to.
    WvIPPortAddr dest_addr;

    if( transproxy && 
    		( (WvIPPortAddr *)src() )->port 
    	     != origin->localaddr().port )

    	dest_addr = origin->localaddr();
    else {
	char *	 err_buf;
	if( transproxy && dest_str == "" ) {
	    fwd.err( "Transparent proxy would form a loop.  Dropped.\n" );
	    delete origin;
	    return;
	}
	if( make_port_addr( dest_addr, dest_str, &err_buf ) == false ) {
    	    fwd.err( err_buf );
	    delete origin;
    	    return;
    	}
    }

    WvFF_TCPConn * dest = new WvFF_TCPConn( dest_addr, origin );
    if( !dest ) {
    	fwd.err( "Connection to %s dropped!\n",
    				*(WvIPPortAddr *) src() );
	delete origin;
	return;
    } else if( dest->isok() == false ) {
    	fwd.err( "Can't forward to %s: %s\n", 
    				*(WvIPPortAddr *) src(), dest->errstr() );
	delete origin;
	return;
    } else {
    	fwd.log( "...forwarding to %s.\n", dest_addr );
    }

    // Now the connection we just made knows about the incoming one, but the
    // incoming one doesn't know about the one we just made.  Fix:
    origin->remote_conn = dest;

    // Tell Fast Forward about both new connections.
    fwd.get_tcp().append( origin, true, "fftcp origin" );
    fwd.get_tcp().append( dest, true, "fftcp dest" );
}


///////////////////////////////////////////////////////////
// WvFF_TCPConn stuff
///////////////////////////////////////////////////////////

unsigned int WvFF_TCPConn::conn_count = 0;

WvFF_TCPConn::WvFF_TCPConn( int open_fd, const WvIPPortAddr& rem )
/****************************************************************/
// protected.  Used by WvFF_TCPListener::accept_ff().
: WvTCPConn( open_fd, rem )
{
    last_xmit   = time( NULL );
    remote_conn = NULL;
    inc_num_conns();
}

WvFF_TCPConn::WvFF_TCPConn( const WvIPPortAddr& rem, WvFF_TCPConn * rem_conn )
/****************************************************************************/
// public.  Used above in WvFF_TCPListener::tcp_accept_callback().
: WvTCPConn( rem )
{
    last_xmit   = time( NULL );
    remote_conn = rem_conn;
    inc_num_conns();
}

WvFF_TCPConn::~WvFF_TCPConn()
/***************************/
{
    if (remote_conn)
    {
	remote_conn->remote_conn = NULL;
	remote_conn->flush_then_close(600*1000); // flush for up to 10 minutes
    }
    dec_num_conns();
}

bool WvFF_TCPConn::pre_select(SelectInfo &si)
/*********************************************/
{
    bool oldr = si.wants.readable;
    bool retval;
    
    // don't select for readable if the remote stream isn't writable; this
    // lets TCP do its buffering properly.
    if (remote_conn && remote_conn->outbuf.used())
	si.wants.readable = false;
    
    retval = WvTCPConn::pre_select(si);
    
    si.wants.readable = oldr;
    return retval;
}

void WvFF_TCPConn::execute()
/*************************************************************/
// Standard stream callback function.  We receive something from s, and we
// need to pipe it through to the remote stream.
{
    char	    buf[ 2048 ];
    int		    num_read;
    
    WvTCPConn::execute();
    
    if( select( 0 ) ) {
	num_read = read( buf, sizeof( buf ) );
	last_xmit = time( NULL );
	if (remote_conn)
	{
	    remote_conn->write( buf, num_read );
	    remote_conn->last_xmit = last_xmit;
	}
    }
}

